#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import importlib.util


polars_not_installed_message = (
    "Polars package not found. "
    "If you want to use Polars with Hopsworks you can install the corresponding extras "
    """`pip install hopsworks[polars]` or `pip install "hopsworks[polars]"` if using zsh. """
    "You can also install polars directly in your environment e.g `pip install polars`. "
    "You will need to restart your kernel if applicable."
)

HAS_POLARS: bool = importlib.util.find_spec("polars") is not None
